#!/usr/bin/perl

##############################################################################
#
# convert_mailboxes.pl
#
# v2.5.3
#
# Advanced hybrid mailbox conversion
#
# Copyright © 2013 Niversoft Idées Logicielles
# info@niversoft.com
#
# This code is provided AS IS, without any certification that it will even run
#
# This script is able to convert mailboxes between mbox, mdir and mslc (CGP6) formats.
# Running the script without parameter will output full commmand line help.
#
# A backup must always be created before running this script.
#
# While this script is free, it has required a lot of time and some experience
# with previous similar scripts to create. If this script is proven useful, or
# if support is required, donation can be send or support time purchased by
# visiting http://www.niversoft.com/purchase, or directly through PayPal to
# billing@niversoft.com.
#
#
# KNOWN LIMITATIONS:
#
# - If you convert a single mailbox or a single account, the Default Mailbox
#   Type setting is changed for the WHOLE domain. Note that if you convert a
#   single account or a single whole domain, other domains are always left
#   untouched.
#
# - This script creates new files under the account it runs on. Make sure you
#   are running it as the user CGP runs on, and with the proper umask value.
#   On Windows, running it elevated is usually fine.
#
##############################################################################

use 5.012_000; # Required for 2038+ date support.
               # Emails arrive with all sorts of wrong dates,
               # one doesn't want the script to crash on them.

use strict;
use Mail::IMAPClient;
use Getopt::Long qw/HelpMessage/;
use Time::Local;
use File::Path qw(make_path remove_tree);

my @mailboxes;
my @accounts;
my @domains;
my $convert_all;
my $convert_file;
my $convert_mode;
my $master_account = "postmaster";
my $master_password;
my $pwd_port = 106;
my $server_address = "127.0.0.1";
my $imap_port = 143;
my $offline_mode;
my $cgp_base = ".";
my $domains_subdir = "Domains";
my $help;
my %accounttree;
my %Months = (Jan=>1,Feb=>2,Mar=>3,Apr=>4,May=>5,Jun=>6,Jul=>7,Aug=>8,Sep=>9,Oct=>10,Nov=>11,Dec=>12);
my @Months = qw/Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec/;
my @WeekDays = qw/Sun Mon Tue Wed Thu Fri Sat/;
my $mbox_size_limit_kb = 0x200000;

my @err = ();
$SIG{'__WARN__'} = sub { push @err, "*** $_[0]" };
my $ok = GetOptions
(  "mailbox:s"        => \@mailboxes,
   "account:s"        => \@accounts,
   "domain:s"         => \@domains,
   "all"              => \$convert_all,
   "file=s"           => \$convert_file,
   "mode=s"           => \$convert_mode,
   "offline"          => \$offline_mode,
   "masteraccount=s"  => \$master_account,
   "masterpassword=s" => \$master_password,
   "server=s"         => \$server_address,
   "cgpbase=s"        => \$cgp_base,
   "domains:s"        => \$domains_subdir,
   "pwdport=i"        => \$pwd_port,
   "imapport=i"       => \$imap_port,
   "help|?"           => \$help,
);

$SIG{'__WARN__'} = "DEFAULT";

$master_account = lc($master_account);

$ok = 0 if (defined $help);
if ($ok)
{
   push (@err, "*** Error: masterpassword and offline mode are mutually exclusive"                                         ) if ( defined $master_password &&  defined $offline_mode);
   push (@err, "*** Error: masterpassword is not defined"                                                                  ) if (!defined $convert_file && !defined $master_password && !defined $offline_mode);
   push (@err, "*** Error: No target specified. \n*** Please use at least one of --mailbox, --account, --domain, --all"    ) unless (defined $convert_file || defined $convert_all || scalar @mailboxes || scalar @accounts || scalar @domains);
   push (@err, "*** Error: The conversion mode is not specified. \n*** Please use --mode=mbox, --mode=mdir or --mode=mslc" ) unless ($convert_mode eq "mdir" || $convert_mode eq "mbox" || $convert_mode eq "mslc");
   push (@err, "*** Error: Unable to find Communigate Pro Base directory. \n*** Please specify --cgpbase."                 ) unless (defined $convert_file || defined FindCGBase());
   push (@err, "*** Error: DomainsDir can't be empty in offline mode. \n*** Please specify --domainsdir."                  ) unless ( !$offline_mode || ($domains_subdir ne "") && -d "$cgp_base/$domains_subdir" );
   push (@err, "*** Error: Direct file conversion can't be used with any option besides --mode."                           ) if ( defined $convert_file && (defined $convert_all || scalar @mailboxes || scalar @accounts || scalar @domains || defined $master_password || defined $offline_mode ));
}

unless ($ok && !scalar @err)
{
   my $err = join "\n", @err;

   if (length $err)
   {
      print "\n$err\n";
      if (-t STDOUT)
      {
         print "\n<Enter> to print usage and exit";
         my $a = <>;
         print "\n";
      }
   }

print <<USAGE;
convert_mailboxes.pl:  Convert one or several mailboxes between mbox, mdir
                       and mslc or vice-versa.

Usage: convert_mailboxes.pl --mailbox=account\@domain/mailbox
                            --account=account\@domain
                            --domain=domain
                            --all
                            --mode=[mbox|mdir|mslc]
                           [--offline]
                           [--server=127.0.0.1]
                           [--pwdport=106]
                           [--imapport=143]
                           [--cgpbase=.|/var/Communigate]
                           [--domains=Domains]
                           [--masteraccount=postmaster]
                           [--masterpassword=password]
                           [--help]

   This script requires Communigate Pro. mbox and mdir formats require CGP 5.1
   or further version. mslc format requires CGP 6.0 or further version.
   This script must run as root.

   INBOXes are converted using IMAP. Other mailboxes are converted directly
   using the filesystem.

   Prior to converting a mailbox, the account\'s password is disabled and any
   existing session is killed, to prevent any external access during the
   conversion. Accesses are restored afterwards.

   Because of this, the master account cannot be converted. To convert the
   postmaster account or to really convert all accounts at once, a temporary
   account with the "Can Do Everything" or the "Can Modify This Domain and
   its Accounts Settings" right should be used as master account.

   Parameters:

   --mailbox, --account, --domain:
               The mailbox, account or domain to convert.
               These settings are cumulative, multiple domains, accounts or
               mailboxes to convert can be specified. An backup of all data
               marked for conversion should have been performed prior to
               running this tool, as it does not have a rollback feature.

               mailbox format: ~account\@domain/mailbox
               account format: account\@domain
               domain format:  domain

   --all       Converts all mailboxes of all accounts of all domains. Requires
               a masteraccount with the "Can Do Everything" right.

   --mode      Target mailbox type, either mbox (TextMailbox), mdir
               (MaildirMailbox) or mslc (SlicedMailbox). Mailboxes already in
               the correct format are not modified.

   --offline   Offline mode only. CLI and IMAP (server, pwdport, imapport)
               are not used. Rights are not checked. Communigate Pro must not
               to be running.

   --file      Special offline mode where only one specific mailbox is
               converted. Use --mode and --file="full filesystem mailbox path"

   --server    IP address of the server for CLI connections

   --pwdport   PWD port, usually 106 or 8106

   --imapport  IMAP port, usually 143. Secure IMAP is not implemented by this
               tool yet.

   --cgpbase   Path to CommunigatePro base directory. if not specified, the
               script will verify if the cwd is a CGP base, then try
               /var/Communigate.

   --domains   Domains subdirectory. Default is "Domains", and is valid for
               most CGP installations. Specify only if your subdomains are
               stored in a different subdirectory, such as "SharedDomains"
               for cluster installations.

   --masteraccount, --masterpassword
               Username and password for the master account which defaults to
               "postmaster\@maindomain". The masteraccount must have either the
               "Can Do Everything" server right, or at least the following
               domain rights for all of the specified domains:
               - Can Modify This Domain and its Accounts Settings
               - Can Access All Mailboxes
               - Basic Settings
               - CommuniGate Password
               - OS Password
               - Mail Storage Limit
               - Mailbox Limit
               - New Mailbox Format

   --help      Shows this help.

USAGE
   exit;

}

$| = 1;

if (defined $convert_file)
{
   print "Converting $convert_file to $convert_mode\n";
   boxconvert($convert_file, $convert_mode);
   exit;
}

my %mailboxes;
my $total_mailboxes = 0;

if (defined $offline_mode)
{
   print "Please stop Communigate Pro before starting the script in offline mode\n" and exit if (Ask("Is Communigate Pro running", 1));

   %mailboxes = map {$_ => 1} @mailboxes;
   my %accounts  = map {$_ => 1} @accounts;

   print "Unable to read Main.settings. Verify cgpbase is set correctly.\n" and exit unless open SETTINGS, "$cgp_base/Settings/Main.settings";
   my $settings = join "", (<SETTINGS>);
   close SETTINGS;
   my $MainDomain = $1 if ($settings =~ /DomainName = ([\w_\-\.]+)/);
   print "Unable to find the Main Domain name. Verify cgpbase is set correctly and cgpbase/Settings/Main.settings is valid.\n" and exit unless(defined $MainDomain);


   my %loc = ();

   if (defined $convert_all)
   {
      print "Enumerating domains\n";
      @domains = ($MainDomain);
   }
   $loc{$MainDomain} = "$cgp_base/Accounts";
   $loc{$MainDomain} =~ s/\\/\//g;
   $loc{$MainDomain} =~ s/\/\//\//g;

   opendir DIR, "$cgp_base/$domains_subdir/";
   foreach my $dom (readdir DIR)
   {
      next if ($dom =~ /^\./);
      if (-d "$cgp_base/$domains_subdir/$dom" && ($dom =~ /^[\w_\-\.]+$/) && $dom ne "Settings" )
      {
         if (defined $convert_all)
         {
            push @domains, $dom;
         }
         $loc{$dom} = "$cgp_base/$domains_subdir/$dom";
         $loc{$dom} =~ s/\\/\//g;
         $loc{$dom} =~ s/\/\//\//g;
      }
   }

   if (scalar @domains)
   {
      print "Enumerating accounts\n";
      foreach my $dom (@domains)
      {
         $accounttree{$dom} = {};
         BuildAccountTree($accounttree{$dom}, $loc{$dom});
         foreach my $acc (keys %{$accounttree{$dom}})
         {
            $accounts{"$acc\@$dom"} = 1;
         }
      }
   }

   @mailboxes = keys %mailboxes;
   %mailboxes = ();

   my $mode_to_convert_from = $convert_mode eq "mdir" ? "(mbox|mslc)"
                            : $convert_mode eq "mbox" ? "(mdir|mslc)"
                            : "(mdir|mbox)";
   sub EnumMailboxes
   {
      my $hash = shift;
      my $base = shift;
      my $loc = shift;
      opendir DIR, $loc;
      my @entries = readdir(DIR);
      closedir DIR;
      foreach my $entry (@entries)
      {
         next if ($entry =~ /^\./);
         if (-d "$loc/$entry" && ($entry =~ /\.folder$/))
         {
            if ($entry =~ /(.*)\.folder$/)
            {
               EnumMailboxes($hash, ($base ? "$base/$1" : $1), "$loc/$entry");
            }
         }
         elsif ($entry =~ /(.*)\.$mode_to_convert_from$/)
         {
            $hash->{"$base/$1"} = "$loc/$entry";
         }
      }
   }

   $total_mailboxes = 0;
   if (scalar keys %accounts)
   {
      print "Enumerating mailboxes\n";
      foreach my $account (keys %accounts)
      {
         if ($account =~ /^~?([^\@\/]+)\@([^\/]+)$/)
         {
            my ($acc, $dom) = ($1, $2);
            $loc{"$acc\@$dom"} = FindAccountLocation($loc{$dom}, $acc, $dom);

            $mailboxes{$dom} = {} unless (defined $mailboxes{$dom});
            $mailboxes{$dom}->{$acc} = {};

            EnumMailboxes($mailboxes{$dom}->{$acc}, "", $loc{"$acc\@$dom"});

            $total_mailboxes += scalar keys %{$mailboxes{$dom}->{$acc}};
         }
      }
   }


   my $cangetaccountlocation = 1;

   print "Creating processing list\n";
   foreach my $mailbox (@mailboxes)
   {
      if ($mailbox =~ /^~?([^\@\/]+)\@([^\/]+)\/(.*)$/)
      {
         my ($acc, $dom, $mbx) = ($1, $2, $3);

         print "Unknown domain: $dom\n" and exit unless defined $loc{$dom};
         $mailboxes{$dom} = {} unless (defined $mailboxes{$dom});

         unless (defined $mailboxes{$dom}->{$acc})
         {
            $mailboxes{$dom}->{$acc} = {};
            $loc{"$acc\@$dom"} = FindAccountLocation($loc{$dom}, $acc, $dom);
         }

         next if (defined $mailboxes{$dom}->{$acc}->{$mbx});

         # guessing location of mailbox.
         my $path = $loc{"$acc\@$dom"};

         my @parts = split "/", $mbx;
         my $part;
         my $found = 0;
         while ($part = shift @parts)
         {
            $path .= "/" unless ($path =~ /\/$/);
            $path .= $part;
            if (scalar @parts)
            {
               $path .= ".folder";
               print "folder $path not found\n" unless (-d $path);
               last unless (-d $path);
            }
            else
            {
               last if ($convert_mode =~ /mdir|mslc/ ? -d "$path.$convert_mode" : -f "$path.$convert_mode");
               $found = (-f "$path.mbox" and $path .= ".mbox") || (-d "$path.mslc" and $path .= ".mslc") if ($convert_mode eq "mdir");
               $found = (-d "$path.mslc" and $path .= ".mslc") || (-d "$path.mdir" and $path .= ".mdir") if ($convert_mode eq "mbox");
               $found = (-d "$path.mdir" and $path .= ".mdir") || (-f "$path.mbox" and $path .= ".mbox") if ($convert_mode eq "mslc");
               print "mailbox $path not found\n" unless $found;
            }
         }
         next unless $found;

         ++$total_mailboxes;
         $mailboxes{$dom}->{$acc}->{$mbx} = $path;
      }
   }
   @mailboxes = ();
   @accounts = ();
   @domains = ();
}
else
{
   my $cli = GetCLI() or print("\nUnable to login through CLI. Please check your settings. Error message: " . $CGP::ERR_STRING . "\n\n") and exit;
   my $imap = GetImap() or print("\nUnable to login through IMAP. Please check your settings.") and exit;
   $imap->logout();

   my $master_prefs = $cli->GetAccountPrefs($master_account);

   %mailboxes = map {$_ => 1} @mailboxes;
   my %accounts  = map {$_ => 1} @accounts;

   my $to_format = $convert_mode eq "mdir" ? "MailDirMailBox"
                 : $convert_mode eq "mbox" ? "TextMailbox"
                 : "SlicedMailbox";

   if (defined $convert_all)
   {
      print "Enumerating domains\n";
      @domains = @{$cli->ListDomains() or print $cli->getErrMessage and exit };
      $cli->UpdateServerAccountDefaults({"DefaultMailboxType"=>($cli->UpdateClusterAccountDefaults({"DefaultMailboxType"=>$to_format}) ? "default" : $to_format)});
   }

   if (scalar @domains)
   {
      print "Enumerating accounts\n";
      foreach my $domain (@domains)
      {
         my %list = ListAccounts($cli, $domain) or print $cli->getErrMessage and exit;
         @accounts{keys %list} = values %list;
      }
   }

   foreach my $todelete ($master_account, $master_prefs->{"AccountName"})
   {
      if (defined $accounts{$todelete})
      {
         print "Skipping account " . $todelete . "\n";
         delete $accounts{$todelete};
      }
   }

   if (scalar keys %accounts)
   {
      print "Enumerating mailboxes\n";
      foreach my $account (keys %accounts)
      {
         my %list = ListMailboxes($cli, $account);
         @mailboxes{keys %list} = values %list;
      }
   }

   @mailboxes = keys %mailboxes;
   %mailboxes = ();
   $total_mailboxes = 0;

   my %loc = ();
   my $cangetaccountlocation = 1;

   print "Creating processing list and setting DefaultMailboxType\n";
   foreach my $mailbox (@mailboxes)
   {
      if ($mailbox =~ /^~?([^\@\/]+)\@([^\/]+)\/(.*)$/)
      {
         my ($acc, $dom, $mbx) = ($1, $2, $3);

         unless (defined $mailboxes{$dom})
         {
            $mailboxes{$dom} = {};

            if ($convert_all)
            {
               $cli->UpdateAccountDefaults(domain => $dom, settings => {"DefaultMailboxType"=>"default"});
            }
            else
            {
               $cli->UpdateAccountSettings("$acc\@$dom", {"DefaultMailboxType"=>"default"}) or print $cli->getErrMessage() and exit;
               $cli->UpdateAccountDefaults(domain => $dom, settings => {"DefaultMailboxType"=>"default"}) or print $cli->getErrMessage() and exit;
               my $AccountEffectiveSettings = $cli->GetAccountEffectiveSettings("$acc\@$dom");
               if ($AccountEffectiveSettings->{"DefaultMailboxType"} !~ /$to_format/i)
               {
                  $cli->UpdateAccountDefaults(domain => $dom, settings => {"DefaultMailboxType"=>$to_format});
               }
            }

            $loc{$dom} = "$cgp_base/" . $cli->GetDomainLocation($dom);
            $loc{$dom} =~ s/\\/\//g;
            $loc{$dom} =~ s/\/\//\//g;
         }

         unless (defined $mailboxes{$dom}->{$acc})
         {
            $mailboxes{$dom}->{$acc} = {};
            $cli->UpdateAccountSettings("$acc\@$dom", {"DefaultMailboxType"=>"default"});

            my $location = $cli->GetAccountLocation("$acc\@$dom") if ($cangetaccountlocation);
            if (defined $location)
            {
               $loc{"$acc\@$dom"} = $loc{$dom} . "/$location";
            }
            else
            {
               $cangetaccountlocation = 0;
               $loc{"$acc\@$dom"} = FindAccountLocation($loc{$dom}, $acc, $dom);
            }
         }

         # guessing location of mailbox.
         my $path = $loc{"$acc\@$dom"};

         my @parts = split "/", $mbx;
         my $part;
         my $found = 0;
         while ($part = shift @parts)
         {
            $path .= "/" unless ($path =~ /\/$/);
            $path .= $part;
            if (scalar @parts)
            {
               $path .= ".folder";
               last unless (-d $path);
            }
            else
            {
               last if ($convert_mode =~ /mdir|mslc/ ? -d "$path.$convert_mode" : -f "$path.$convert_mode");
               $found = (-f "$path.mbox" and $path .= ".mbox") || (-d "$path.mslc" and $path .= ".mslc") if ($convert_mode eq "mdir");
               $found = (-d "$path.mslc" and $path .= ".mslc") || (-d "$path.mdir" and $path .= ".mdir") if ($convert_mode eq "mbox");
               $found = (-d "$path.mdir" and $path .= ".mdir") || (-f "$path.mbox" and $path .= ".mbox") if ($convert_mode eq "mslc");
            }
         }
         next unless $found;

         ++$total_mailboxes;
         $mailboxes{$dom}->{$acc}->{$mbx} = $path;
      }
   }
   @mailboxes = ();
   @accounts = ();
   @domains = ();
   $cli->Logout();
}

print "$total_mailboxes mailboxes to convert\n";
exit unless $total_mailboxes;

while(1)
{
   my $ask = Ask("Do you have an up-to-date backup (l=list accounts)", 0, {"y"=>1,"n"=>0,"l"=>-1});
   print "Make sure to create a full backup prior to running this script\n" and exit unless ($ask);
   last if ($ask == 1);

   print "\nSelected data:\n";
   foreach my $domain (sort keys %mailboxes)
   {
      print "$domain\n";
      foreach my $account (sort keys %{$mailboxes{$domain}})
      {
         print "   $account (" . (scalar keys %{$mailboxes{$domain}->{$account}}) . " mailboxes)\n";
      }
   }
   print "\n";
}

my $domain_count = scalar keys %mailboxes;
my $domain_index = 0;
foreach my $domain (sort keys %mailboxes)
{
   print "DOMAIN " . ++$domain_index . "/$domain_count - $domain\n";

   my $account_count = scalar keys %{$mailboxes{$domain}};
   my $account_index = 0;
   foreach my $account (sort keys %{$mailboxes{$domain}})
   {
      print "  ACCOUNT " . ++$account_index . "/$account_count - $account\@$domain\n";
      my $fan = "$account\@$domain";
      my $settings;

      if (!$offline_mode)
      {
         my $cli = GetCLI();

         $settings = $cli->GetAccountSettings($fan);

         if (!$cli->UpdateAccountSettings($fan, {UseAppPassword => "NO", UseSysPassword => "NO", MaxAccountSize => "-1", MaxMailboxes => "-1"}))
         {
            print "ERROR: Unable to change password settings, skipping account $fan\n" . $cli->getErrMessage() . "\n";
            $cli->Logout();
            next;
         }
         if (!$cli->KillAccountSessions($fan))
         {
            $cli->UpdateAccountSettings($fan,
               {  UseAppPassword => GetVal($settings, "UseAppPassword")
               ,  UseSysPassword => GetVal($settings, "UseSysPassword")
               ,  MaxAccountSize => GetVal($settings, "MaxAccountSize")
               ,  MaxMailboxes   => GetVal($settings, "MaxMailboxes")
               });
            print "ERROR: Unable to kill acccount sessions, skipping account $fan\n" . $cli->getErrMessage() . "\n";
            $cli->Logout();
            next;
         }
         $cli->Logout();
      }

      my $mailbox_count = scalar keys %{$mailboxes{$domain}->{$account}};
      my $mailbox_index = 0;
      sub inboxfirst($$) { return $_[0] eq "INBOX" ? -1 : ($_[1] eq "INBOX" ? 1 : ($_[0] cmp $_[1])) }
      foreach my $mailbox (sort inboxfirst keys %{$mailboxes{$domain}->{$account}})
      {
         print "    MAILBOX " . ++$mailbox_index . "/$mailbox_count - $mailbox\n";

         my $path = $mailboxes{$domain}->{$account}->{$mailbox};

         if (!$offline_mode && ($mailbox eq "INBOX"))
         {
            my $cli = GetCLI();
            my $acl = $cli->GetMailboxACL($fan, $mailbox);
            $cli->Logout();

            ImapRebuild($fan, $mailbox);

            $cli = GetCLI();
            $cli->SetMailboxACL($fan, $mailbox, $acl);
            $cli->Logout();
         }
         else
         {
            boxconvert($path, $convert_mode);
         }
      }

      if (!$offline_mode)
      {
         my $cli = GetCLI();

         $cli->UpdateAccountSettings($fan,
            {  UseAppPassword => GetVal($settings, "UseAppPassword")
            ,  UseSysPassword => GetVal($settings, "UseSysPassword")
            ,  MaxAccountSize => GetVal($settings, "MaxAccountSize")
            ,  MaxMailboxes   => GetVal($settings, "MaxMailboxes")
            })
            or print "ERROR: Unable to change back settings, this account will remain disabled\n";

         $cli->Logout();
      }
   }
}

sub GetCLI
{
   require CLI;
   return new CGP::CLI({ PeerAddr => $server_address, PeerPort => $pwd_port, login => $master_account, password => $master_password, SecureLogin => 0 });
}

sub GetImap
{
   my $imap = Mail::IMAPClient->new();
   $imap->Fast_io(0);
   $imap->Uid(0);
   $imap->Server($server_address);
   $imap->Port($imap_port);
   $imap->connect or print "Cannot connect to host: $@" and return;
   $imap->User($master_account);
   $imap->Password($master_password);
   $imap->login() or print "Cannot login with username $master_account: $@" and return;
   return $imap;
}

sub GetVal
{
   return defined $_[0]->{$_[1]} ? $_[0]->{$_[1]} : "default";
}

sub ListMailboxes
{
   return map +("$_[1]/$_", 1), (keys %{$_[0]->ListMailboxes(accountName => $_[1])});
}

sub ListAccounts
{
   return map +(lc("$_\@$_[1]"), 1), (keys %{$_[0]->ListAccounts ($_[1])});
}

sub Ask
{
   my $question = shift;
   my $default = shift;
   my $choices = shift;
   $choices = {"y"=>1, "n"=>0} unless defined $choices;

   my @choices;
   foreach my $choice (keys %{$choices})
   {
      push @choices, substr($choices->{$choice} == $default ? uc($choice) : $choice, 0, 1);
   }
   $question .= " [" . join("/", @choices) . "]? ";
   while (1)
   {
      print $question;
      my $resp = <>;
      chomp $resp;
      return $default unless (length $resp);
      $resp = lc(substr($resp,0,1));
      return $choices->{$resp} if (defined $choices->{$resp});
   }
}

sub FindCGBase
{
   foreach ($cgp_base, "/var/Communigate", "c:/Communigate Files", "/cygdrive/c/Communigate Files")
   {
      return($_) if (-f "$_/Settings/Main.settings" && -d "$_/Accounts")
   };
   return undef;
}

sub BuildAccountTree
{
   my $hash = shift;
   my $path = shift;

   opendir DIR, $path ;
   my @sublist = readdir(DIR);
   closedir DIR;

   foreach my $dir (@sublist)
   {
      next if ($dir =~ /^\./);
      if (-d "$path/$dir")
      {
         if ($dir =~ /(.*)\.macnt$/)
         {
            $hash->{$1} = "$path/$dir";
         }
         elsif ($dir =~ /.sub$/)
         {
            BuildAccountTree($hash, "$path/$dir");
         }
      }
   }
}

sub FindAccountLocation
{
   my $domainpath = shift;
   my $accountname = shift;
   my $domain = shift;

   unless (defined $accounttree{$domain})
   {
      $accounttree{$domain} = {};
      BuildAccountTree($accounttree{$domain}, $domainpath);
   }
   return $accounttree{$domain}->{$accountname};
}

sub ImapRebuild
{
   my $account = shift;
   my $mailbox = shift;

   my $mailboxcopy = $mailbox . "_tmp_" . time();

   my $imap = GetImap() or print("Error: Unable to login through IMAP. Skipping mailbox $mailbox.\n") and return;

   sub error_close { print $_[0] and $imap->logout(); return 0; }

   $imap->select("~$account/$mailbox") or return error_close("Error: Unable to select $mailbox. Skipping mailbox.\n");
   $imap->close();
   $imap->rename("~$account/$mailbox", "~$account/$mailboxcopy") or return error_close("Error: Unable to rename $mailbox. Skipping mailbox.\n");

   $imap->create("~$account/$mailbox"); # just to be sure the box is created. No error check.
   $imap->select("~$account/$mailboxcopy") or return error_close("Error: Unable to select $mailboxcopy. Unable to move back messages in $mailbox.\n");

   my $last = $imap->message_count();

   my $size = 0;
   for (my $i = 1; $i <= $last;)
   {
      my $rangelast = ($last < $i + 49) ? $last : ($i + 49);

      my $splitafter = 0;
      if ($convert_mode eq "mbox")
      {
         for my $j ($i..$rangelast)
         {
            my $msgsize = RoundKbUp($imap->size($j));
            $size += $msgsize;
            if ($size > $mbox_size_limit_kb)
            {
               $rangelast = $j - 1;
               $splitafter = 1;
               $size = 0;
            }
         }
      }

      my $range = $imap->Range($i..$rangelast);
      if (!$imap->copy("~$account/$mailbox", $range))
      {
         for (my $j = $i; $j <= $rangelast; ++$j)
         {
            $imap->copy("~$account/$mailbox", $j);
         }
      }
      $imap->delete_message($range);

      if ($splitafter && ($rangelast < $last))
      {
         print "Mailbox too big, splitting.\n";

         my $t = 1;
         ++$t while ($imap->exists("~$account/$mailbox.$t"));
         for my $i ($t-1..1)
         {
            $imap->rename("~$account/$mailbox.$i", "~$account/$mailbox." . ($i + 1));
         }
         $imap->rename("~$account/$mailbox", "~$account/$mailbox.1");

         $imap->create("~$account/$mailbox");
      }

      $i = $rangelast + 1;
   }
   $imap->close();
   $imap->delete("~$account/$mailboxcopy") or return error_close("Error: Unable to remove $mailboxcopy.\n");
   $imap->logout();
   return 1;
}

sub RoundKbUp
{
   my $num = $_[0] + 100;
   return ($num - ($num % 1024)) / 1024 + 1; # ceil()-like function with padding
}


my $wineol = $^O eq 'MSWin32';
sub boxconvert
{
   my $path = shift;
   my $targetmode = shift;
   my $sourcemode;
   if (-d $path)
   {
      $sourcemode = "mdir" if ($path =~ /\.mdir$/);
      $sourcemode = "mslc" if ($path =~ /\.mslc$/);
   }
   elsif (-f $path)
   {
      $sourcemode = "mbox" if ($path =~ /\.mbox$/);
   }
   else
   {
      print "Path not found: $path\n";
      return;
   }

   print "Unknown source path format: $path\n" and return unless (defined $sourcemode);
   print "Path is already in the correct format: $path\n" and return if ($sourcemode eq $targetmode);

   my $sourcedata = InitSource($sourcemode, $path);
   return unless defined($sourcedata);

   my $targetdata = InitTarget($targetmode, $path);
   CompleteSource($sourcemode, $sourcedata, 0) and return unless defined($targetdata);


   my $message = ReadMessage($sourcemode, $sourcedata);
   while (defined $message)
   {
      if (!WriteMessage($targetmode, $targetdata, $message))
      {
         CompleteSource($sourcemode, $sourcedata, 0);
         CompleteTarget($targetmode, $targetdata, 0);
      }
      $message = ReadMessage($sourcemode, $sourcedata);
   }

   CompleteSource($sourcemode, $sourcedata, 1);
   CompleteTarget($targetmode, $targetdata, 1);
}

sub InitSource
{
   my $mode = shift;
   my $path = shift;
   if ($mode eq "mdir")
   {
      opendir DIR, $path;
      my @files = sort {$a <=> $b} readdir(DIR);
      closedir DIR;
      return {"d"=>$path, "f"=>\@files, "i"=>0};
   }
   elsif ($mode eq "mbox")
   {
      my $mbox;
      open $mbox, $path or print "Error reading TextMailbox file, skipping mailbox\n" and return;
      return {"b"=>$path, "f"=>$mbox};
   }
   elsif ($mode eq "mslc")
   {
      opendir DIR, $path;
      my @files = sort {substr($a, 4) <=> substr($b, 4)} readdir(DIR);
      closedir DIR;
      return {"d"=>$path, "fl"=>\@files, "i"=>0};
   }
}

sub InitTarget
{
   my $mode = shift;
   my $path = shift;
   if ($mode eq "mdir")
   {
      my $mdirname = $path;
      $mdirname =~ s/(mslc|mbox)$/mdir/;
      mkdir $mdirname or print "Error creating MaildirMailbox $mdirname, skipping mailbox\n" and return;
      return {"d"=>$mdirname};
   }
   elsif ($mode eq "mbox")
   {
      my $mboxname = $path;
      my $mbox;
      $mboxname =~ s/(mdir|mslc)$/mbox/;
      open $mbox, ">", $mboxname or print "can't open output file, skipping $path" and return;
      return {"b"=>[$mboxname], "f"=>$mbox, "s"=>0};
   }
   elsif ($mode eq "mslc")
   {
      my $mslcname = $path;
      $mslcname =~ s/(mdir|mbox)$/mslc/;
      mkdir $mslcname or print "Error creating SlicedMailbox $mslcname, skipping mailbox\n" and return;
      return {"d"=>$mslcname, "fi"=>0, "f"=>undef, "n"=>0 };
   }
}

sub CompleteSource
{
   my $mode = shift;
   my $data = shift;
   my $success = shift;

   if ($mode =~ /mdir|mslc/)
   {
      remove_tree($data->{"d"}) if ($success);
   }
   elsif ($mode eq "mbox")
   {
      close $data->{"f"};
      unlink $data->{"b"} if ($success);
   }
   return 1;
}

sub CompleteTarget
{
   my $mode = shift;
   my $data = shift;
   my $success = shift;

   if ($mode =~ /mdir|mslc/)
   {
      remove_tree($data->{"d"}) unless ($success);
   }
   elsif ($mode eq "mbox")
   {
      close $data->{"f"};
      unless ($success)
      {
         foreach my $box (@{$data->{"b"}})
         {
            unlink $box;
         }
      }
   }

   return 1;
}

sub ReadMessage
{
   my $mode = shift;
   my $data = shift;
   my $message;
   if ($mode eq "mdir")
   {
      while (1)
      {
         my $file = $data->{"f"}->[$data->{"i"}++];
         return unless defined($file);
         next if $file =~ /\./;

         if ($file =~ /^(\d+)(?:O\d+)?-([A-Z_]+)-(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})-(\d+)/i)
         {
            my ($id, $flags, $year, $mon, $mday, $h, $m, $s) = ($1, $2, $3, $4, $5, $6, $7, $8);
            next if (substr($flags, 4, 1) eq "P"); # skipping items marked as deleted
            $file = $data->{"d"} . "/" . $file;
            if (length $flags == 8)
            {
               $flags .= "_____";
               ($s,$m,$h,$mday,$mon,$year) = localtime(timelocal($s,$m,$h,$mday,$mon - 1,$year));
               $year += 1900;
               $mon++;
               $mon = "0$mon" if ($mon < 10);
               $mday = "0$mday" if ($mday < 10);
               $h = "0$h" if ($h < 10);
               $m = "0$m" if ($m < 10);
               $s = "0$s" if ($s < 10);
            }

            my @lines;
            open(MDIR, $file) or print "Can't open input file, skipping mailbox" and return ;
            foreach my $line (<MDIR>)
            {
               $line =~ s/[\r\n]//g;
               $line =~ s/^(>*From )/>$1/;
               push @lines, $line;
            }
            close MDIR;
            push @lines, "";
            return {"id"=>$id - 0, "flags"=>$flags, "year"=>$year, "mon"=>$mon, "mday"=>$mday, "h"=>$h, "m"=>$m, "s"=>$s, "lines"=>\@lines };
         }
      }
   }
   elsif ($mode eq "mbox")
   {
      my ($id, $flags, $year, $mon, $mday, $h, $m, $s);
      my $mbox = $data->{"f"};
      my $line = defined $data->{"l"} ? $data->{"l"} : <$mbox>;
      my @lines;
      while ($line)
      {
         $line =~ s/[\r\n]//g;
         if ($line =~ /^From <>\(/)
         {
            $data->{"l"} = $line;
            return {"id"=>$id - 0, "flags"=>$flags, "year"=>$year, "mon"=>$mon, "mday"=>$mday, "h"=>$h, "m"=>$m, "s"=>$s, "lines"=>\@lines } if (defined $id);

            undef $id;
            if ($line =~ /^From <>\(([A-Z_]+)-(\d+)(?:-\d+)?\)/)
            {
               ($flags, $id) = ($1, $2);

               if ($line =~ / \w+ (\w+) (\d+) (\d+)\:(\d+)\:(\d+) (\d+)[_\s\0]*$/)
               {
                  ($mon, $mday, $h, $m, $s, $year) = ($1, $2, $3, $4, $5, $6);
                  $mon = $Months{$mon};
                  $mon = "0$mon" if ($mon < 10);
               }
               elsif ($line =~ / (\d+)-(\d+)-(\d+)_(\d+)\:(\d+)\:(\d+)[_\s\0]*$/)
               {
                  ($mday, $mon, $year, $h, $m, $s) = ($1, $2, $3, $4, $5, $6)
               }
               else
               {
                  undef $id;
               }
               $flags = substr($flags, 0, 13);
               undef $id if (substr($flags, 4, 1) eq "P");
            }
         }
         elsif (defined $id)
         {
            $line =~ s/^>(>*From )/$1/;
            push @lines, $line;
         }
         $line = <$mbox>;
      }
      push @lines, "";
      undef $data->{"l"};
      return {"id"=>$id - 0, "flags"=>$flags, "year"=>$year, "mon"=>$mon, "mday"=>$mday, "h"=>$h, "m"=>$m, "s"=>$s, "lines"=>\@lines } if (defined $id);
   }
   elsif ($mode eq "mslc")
   {
      unless (defined($data->{"f"}))
      {
         my $i = $data->{"i"};
         $i++ while (defined $data->{"fl"}->[$i] && $data->{"fl"}->[$i] !~ /data\d+/);

         $data->{"i"} = $i;
         return unless defined $data->{"fl"}->[$i];

         my $filename = $data->{"d"} . "/" . $data->{"fl"}->[$i];

         my $file;
         open $file, $filename or print "can't open output file, skipping " . $data->{"d"} and return;
         $data->{"f"} = $file;
      }

      my ($id, $flags, $year, $mon, $mday, $h, $m, $s);
      my $mbox = $data->{"f"};
      my $line = defined $data->{"l"} ? $data->{"l"} : <$mbox>;
      my @lines;
      while ($line)
      {
         $line =~ s/[\r\n]//g;
         if ($line =~ /^\x01\x01\x01/)
         {
            $data->{"l"} = $line;
            if (defined $id)
            {
               push @lines, "";
               return {"id"=>$id - 0, "flags"=>$flags, "year"=>$year, "mon"=>$mon, "mday"=>$mday, "h"=>$h, "m"=>$m, "s"=>$s, "lines"=>\@lines };
            }

            undef $id;
            if ($line =~ /^\x01\x01\x01\x30 \d+ (\d+) (\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d) ([A-Z_]+)/)
            {
               ($id, $year, $mon, $mday, $h, $m, $s, $flags) = ($1, $2, $3, $4, $5, $6, $7, $8, $9);

               $flags = substr($flags, 0, 13);
               undef $id if (substr($flags, 4, 1) eq "P");
            }
         }
         elsif (defined $id)
         {
            push @lines, $line;
         }
         $line = <$mbox>;
      }
      undef $data->{"l"};
      undef $data->{"f"};
      ++$data->{"i"};
      push @lines, "";
      return {"id"=>$id - 0, "flags"=>$flags, "year"=>$year, "mon"=>$mon, "mday"=>$mday, "h"=>$h, "m"=>$m, "s"=>$s, "lines"=>\@lines } if (defined $id);
   }
   return undef;
}

sub WriteMessage
{
   my $mode = shift;
   my $data = shift;
   my $message = shift;

   if ($mode eq "mdir")
   {
      my ($id, $flags, $year, $mon, $mday, $h, $m, $s, $lines) = ($message->{"id"}, $message->{"flags"}, $message->{"year"}, $message->{"mon"}, $message->{"mday"}, $message->{"h"}, $message->{"m"}, $message->{"s"}, $message->{"lines"});
      my $filename = "$id-$flags-$year$mon$mday$h$m$s-";
      my $numlines = scalar @$lines - 1;
      if ($numlines)
      {
         open(MDIR, ">",  $data->{"d"} . "/$filename$numlines") or print "Error writing message in " . $data->{"d"} . ", skipping mailbox\n" and return 0;
         print MDIR join("\n", @$lines);
         close MDIR;
      }
      return 1;
   }
   elsif ($mode eq "mbox")
   {
      my ($id, $flags, $year, $mon, $mday, $h, $m, $s, $lines) = ($message->{"id"}, $message->{"flags"}, $message->{"year"}, $message->{"mon"}, $message->{"mday"}, $message->{"h"}, $message->{"m"}, $message->{"s"}, $message->{"lines"});
      $flags = "${flags}_";
      $id = "0$id" while (length $id < 12);
      my $date = "$mday-$mon-${year}_$h:$m:${s}_";

      my $first = $data->{"s"} ? "\n" : "";
      $lines = "From <>($flags-$id) $date\n" . join("\n", @$lines);
      $data->{"s"} = 1;

      my $filesize = (length $lines) + ($wineol ? ($lines =~ tr/\n//) : 0) + 1;
      if ((tell $data->{"f"}) + $filesize >= $mbox_size_limit_kb * 1024) # too big, splitting in multiple parts.
      {
         close $data->{"f"};

         print "Mailbox too big, splitting\n";

         my $newname = $data->{"b"}->[0];
         $newname =~ s/\.mbox$//;

         my $t = 1;
         ++$t while (-f "$newname.$t.mbox");
         push @{$data->{"b"}}, "$newname.$t.mbox";
         if ($t > 1)
         {
            for my $i (reverse(1..$t-1))
            {
               rename("$newname.$i.mbox", "$newname." . ($i + 1) . ".mbox") or print "Error renaming $newname.$i.mbox, skipping mailbox\n" and return 0;
            }
         }
         rename ($data->{"b"}->[0], "$newname.1.mbox") or print "Error renaming " . $data->{"b"}->[0] . ", skipping mailbox\n" and return 0;
         open ($data->{"f"}, ">" . $data->{"b"}->[0]) or print "Error writing message in " . $data->{"b"}->[0] . ", skipping mailbox\n" and return 0;
         $first = "";
      }

      print {$data->{"f"}} $first if(length $first);
      print {$data->{"f"}} $lines;
      return 1;
   }
   elsif ($mode eq "mslc")
   {
      unless (defined $data->{"f"})
      {
         open($data->{"f"}, ">",  $data->{"d"} . "/data" . (++$data->{"fi"})) or print "Error writing message in " . $data->{"d"} . ", skipping mailbox\n" and return 0;
      }

      my ($id, $flags, $year, $mon, $mday, $h, $m, $s, $lines) = ($message->{"id"}, $message->{"flags"}, $message->{"year"}, $message->{"mon"}, $message->{"mday"}, $message->{"h"}, $message->{"m"}, $message->{"s"}, $message->{"lines"});
      $id = "0$id" while (length $id < 10);

      my $lines = "\x01\x01\x010 $id $id ${year}${mon}${mday}${h}${m}${s} ${flags}_____________\n" . join("\n", @$lines);
      my $filesize = (length $lines) + ($wineol ? ($lines =~ tr/\n//) : 0);
      if ((tell $data->{"f"}) + $filesize >= 104857600 || $data->{"n"} > 1000) # too big, splitting in multiple parts.
      {
         close $data->{"f"};
         open($data->{"f"}, ">",  $data->{"d"} . "/data" . (++$data->{"fi"})) or print "Error writing message in " . $data->{"d"} . ", skipping mailbox\n" and return 0;
         $data->{"n"} = 0;
      }
      ++$data->{"n"};
      print {$data->{"f"}} $lines;
      return 1;
   }

   return 1;
}
